/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source 
 * code directory.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/select.h>
#include <stdlib.h>
#include <errno.h>
#include <netinet/in.h>
#include "runtime.h"
#include "ipfunc.h"

static int ipv4_range_to_bitcount(struct ipv4_range *range);

/* Converts a string in dotted quotted form to an ipv4_addr structure.
 * The following formats are accepted:
 * a.b.c.d
 * a.b.c.d/bitnum
 * a.b.c.d/e.f.g.h
 *
 * The bitnum field corisponds to the number of 1 bits in the netmask.
 * eg /24 means 24 bits are 1, which works out to 255.255.255.0
 */
int str_to_ipv4(const char *str, struct ipv4_range *ans)
{
	struct ipv4_range range;
	const char *cp;
	char *next;
	int bit;
	int tmp,i;

	/* We don't polute the ipv4 return structure with an invalid
	 * value, hence we use our on working copy (ip) and copy
	 * the result to ipv4 if we succeed
	 */

	/* Believe it or not, its easier to understand and work with
	 * if you unroll the loop
	 */
	cp = str;
	bit = strtol(cp,&next,10);
	if (bit < 0 || bit > 255 || cp == next || *next != '.') 
		return -1;
	range.ip.addr.sep.a = bit;

	cp = next+1;
	bit = strtol(cp,&next,10);
	if (bit < 0 || bit > 255 || cp == next || *next != '.') 
		return -1;
	range.ip.addr.sep.b = bit; 

	cp = next+1;
	bit = strtol(cp,&next,10);
	if (bit < 0 || bit > 255 || cp == next || *next != '.') 
		return -1;
	range.ip.addr.sep.c = bit; 

	cp = next+1;
	bit = strtol(cp,&next,10);
	if (bit < 0 || bit > 255 || cp == next) 
		return -1;
	range.ip.addr.sep.d = bit; 

	if (*next == 0) {
		ans->ip.addr.ulong = range.ip.addr.ulong;
		ans->nm.addr.ulong = 0xFFFFFFFF;
		return 0;
	}
	if (*next != '/') 
		return -1;

	cp = next+1;
	bit = strtol(cp,&next,10);
	if (cp == next)
		return -1;
	if (*next == 0 && bit >= 0 && bit < 33) {
		ans->ip.addr.ulong = range.ip.addr.ulong;
		/* TODO 
		 * I know this can be done much faster, but still remain
		 * portable accross endianness, please let me know :)
		 */
		/* if bit == 16: res is 255.255.0.0 or 0xFFFF0000 */
		/* if bit == 8:  res is 255.0.0.0   or 0xFF000000 */
		tmp = bit/8;
		range.nm.addr.ulong = 0;
		switch(tmp) {
		case 4: range.nm.addr.arr[3] = 255;
		case 3: range.nm.addr.arr[2] = 255;
		case 2: range.nm.addr.arr[1] = 255;
		case 1: range.nm.addr.arr[0] = 255;
		}
		i = tmp;
		tmp = bit % 8;
		if (i != 4 && tmp != 0) { /* No remainder, no left over bits */
			/* Bit: 7 = 11111110 or 0xFE
			 * Bit: 1 = 10000000 or 0x10
			 */
			range.nm.addr.arr[i] = (unsigned char)(255 << (8-tmp));
		}
		ans->nm.addr.ulong = range.nm.addr.ulong;
		return 0;
	} else if (*next != '.') {
		return -1;
	}
	if (bit < 0 || bit > 255) 
		return -1;
	range.nm.addr.sep.a = bit;

	cp = next+1;
	bit = strtol(cp,&next,10);
	if (bit < 0 || bit > 255 || cp == next || *next != '.') 
		return -1;
	range.nm.addr.sep.b = bit;

	cp = next+1;
	bit = strtol(cp,&next,10);
	if (bit < 0 || bit > 255 || cp == next || *next != '.') 
		return -1;
	range.nm.addr.sep.c = bit;

	cp = next+1;
	bit = strtol(cp,&next,10);
	if (bit < 0 || bit > 255 || *next != 0) 
		return -1;
	range.nm.addr.sep.d = bit;

	ans->ip.addr.ulong = range.ip.addr.ulong;
	ans->nm.addr.ulong = range.nm.addr.ulong;

	return 0;
}

/* Does the same as above, except expects only an IPv4 address.
 */
int str_to_ipv4_addr(const char *str, struct ipv4_addr *ipv4_a)
{
	struct ipv4_range r;
	if (str_to_ipv4(str,&r) == -1)
		return -1;
	ipv4_a->addr.ulong = r.ip.addr.ulong;
	return 0;
}


/* The following converts an ipv4 structure to a string.
 * The output depends on the style parameter and the netmask setting.
 * of the ipv4 address
 * IPV4_IP_ONLY
 *    Output is always in the form:
 *    a.b.c.d
 * 
 * IPV4_IP_OPT_FULL_NM
 *    If netmask is anything but all 1's (eg /255.255.255.255) then
 *    the output is in the form:
 *    a.b.c.d/e.f.g.h
 *    otherwise its simply: a.b.c.d
 *    
 * IPV4_IP_OPT_SMART_NM
 *    If the netmask is anything but all 1's (eg /255.255.255.255) then
 *    and the netmask is bit divisible, the output is in the form:
 *    a.b.c.d/bitnum
 *    If the netmask is wierd (eg: /123.211.12.1) then the output is
 *    in the form:
 *    a.b.c.d/e.f.g.h
 *    Otherwise, the output is in the form:
 *    a.b.c.d
 *    
 * IPV4_IP_WITH_SMART_NM   
 *    If the netmask is bit divisible, then the output is in the form:
 *    a.b.c.d/bitnum
 *    otherwise its:
 *    a.b.c.d/e.f.g.h
 *
 * IPV4_IP_WITH_FULL_NM
 *    The output is in the form:
 *    a.b.c.d/e.f.g.h
 * 
 * IPV4_FULL_NM_ONLY
 *    The output is in the form:
 *    e.f.g.h
 *
 * IPV4_SMART_NM_ONLY
 *    If the netmask is bit divisible, then the output is in the form:
 *    bitnum
 *    Otherwise its in the form:
 *    e.f.g.h
 */
int ipv4_to_str(struct ipv4_range *ipv4, char *buffer, 
		size_t bufflen, int style)
{
	char *out;
	size_t len;
	int r;
	out = buffer;
	
	/* Write IP address */
	switch(style) {
	case IPV4_IP_ONLY:
	case IPV4_IP_OPT_FULL_NM:
	case IPV4_IP_OPT_SMART_NM:
	case IPV4_IP_WITH_SMART_NM:   
	case IPV4_IP_WITH_FULL_NM:
		if (ipv4_addr_to_str(&ipv4->ip,out,bufflen) == -1)
			return -1;
		len = strlen(out);
		bufflen -= len;
		out+=len;
	}
	if (bufflen <= 0)
		return 0;

	switch(style) {
	case IPV4_IP_ONLY:
		return 0;
	/* Add the slash if needed */
	case IPV4_IP_OPT_FULL_NM:
	case IPV4_IP_OPT_SMART_NM:
		/* Check if we need to display the netmask */
		if (ipv4->nm.addr.ulong == 0xFFFFFFFF)
			return 0;
	case IPV4_IP_WITH_SMART_NM:   
	case IPV4_IP_WITH_FULL_NM:
		*out = '/';
		out++;
		*out = 0;
		break;
	}

	switch(style) {
	/* Do Smart Netmask */
	case IPV4_IP_OPT_SMART_NM:
	case IPV4_SMART_NM_ONLY:
	case IPV4_IP_WITH_SMART_NM:   
		r = ipv4_range_to_bitcount(ipv4);
		if (r != -1) {
			snprintf(out,bufflen,"%d",r);
			len = strlen(out);
			bufflen-=len;
			out+=len;
			break;
		}
		/* fall through to full */
	/* Do FULL Netmask */
	case IPV4_IP_OPT_FULL_NM:
	case IPV4_FULL_NM_ONLY:
	case IPV4_IP_WITH_FULL_NM:
		if (ipv4_addr_to_str(&ipv4->nm,out,bufflen) == -1)
			return -1;
		len = strlen(out);
		bufflen -= len;
		out+=len;
	
		break;
	
	}
	return 0;
}

int ipv4_legal(struct ipv4_range *range)
{
	int i,must_zero=0;
	for(i = 0; i < 4; i++) {
		if (must_zero && range->nm.addr.arr[i] != 0)
			return -1;
		switch(range->nm.addr.arr[i]) {
		case 255: break;
		case 254:  
		case 252:  
		case 248:  
		case 240:  
		case 224:  
		case 192:  
		case 128:  
		case 0: must_zero = 1; break; 
		default:
			return -1;
		}
		
	}
	return 0;
}

int ipv4_range_to_bitcount(struct ipv4_range *range)
{
	int r = 0,i;
	/* Theres a better way to do this, I just wish I knew
	 * what it was
	 */
	for(i = 0; i < 4; i++) {
		switch(range->nm.addr.arr[i]) {
		case 255: r+=8; break;
		case 254: r+=7; return r;
		case 252: r+=6; return r;
		case 248: r+=5; return r;
		case 240: r+=4; return r;
		case 224: r+=3; return r;
		case 192: r+=2; return r;
		case 128: r+=1; return r;
		case 0: return r;
		default: return -1;
		}
	}
	return r;
}

	

int ipv4_addr_to_str(struct ipv4_addr *a, char *buffer,
		size_t bufflen)
{
	snprintf(buffer,bufflen,"%d.%d.%d.%d",
			a->addr.sep.a,
			a->addr.sep.b,
			a->addr.sep.c,
			a->addr.sep.d);
	return 0;
}

/* Returns 0 if the ip address is a loopback address an the netmask
 * is greater than 8.
 * eg a ipv4_comp() call will match this address to something in the
 * 127.0.0.0 region. (may not match 127.0.0.1 if for example, the
 * ipv4 struct was 127.0.0.5/32)
 * Returns -1 otherwise
 */
int ipv4_is_loopback(struct ipv4_range *ipv4)
{
	if (ipv4->ip.addr.sep.a == 127 &&
			ipv4->nm.addr.sep.a == 255)
		return 0;
	return -1;
}

/* Returns 0 if the ipv4 address is a wildcard (0.0.0.0/0.0.0.0)
 */
int ipv4_all_balls(struct ipv4_range *ipv4)
{
	if (ipv4->ip.addr.ulong == 0 &&
			ipv4->nm.addr.ulong == 0)
		return 0;
	return -1;
}
/* Returns 0 if the ipv4 address is a smurf (255.255.255.255)
 */
int ipv4_smurf(struct ipv4_addr *ipv4)
{
	if (ipv4->addr.ulong == 0xFFFFFFFF)
		return 0;
	return -1;
}

/* ipv4_comp(address,match)
 * This function compares an ipv4_addr address, and a ipv4_addr
 * address and its netmask.
 * The netmask attribute of the address parameter is ignored.
 * The netmask atrribute of the match parameter is NOT ignored.
 * Returns 0 on a match, and -1 otherwise.
 *
 * Examples: (Assume the strings are actually ipv4_addr structures)
 * ipv4_comp("1.2.3.4","0.0.0.0/0") = 0
 * ipv4_comp("1.2.3.4","255.255.255.255/32") = -1
 * ipv4_comp("1.2.3.4","255.255.255.255/0") = 0 (NOTE: the netmask!)
 * ipv4_comp("192.168.1.2","192.168.0.0/16") = 0
 * ipv4_comp("192.168.1.2","192.168.0.0/24") = -1 (NOTE: the netmask!)
 * ipv4_comp("192.168.1.10","192.168.1.0/255.255.255.248") = 0
 */
int ipv4_comp(struct ipv4_addr *address, struct ipv4_range *match)
{
	int i;
	unsigned char a,m_a,m_n;
	/* We'll let the compiler unroll this :) */
	for(i = 0; i < 4; i++) {
		a = address->addr.arr[i];
		m_a = match->ip.addr.arr[i];
		m_n = match->nm.addr.arr[i];
		if ((m_a & m_n) != (a & m_n))
			return -1;
	}
	return 0;
}

/* Converts an in_addr to an ipv4_addr */
int in_addr_to_ipv4(struct in_addr *in, struct ipv4_addr *addr)
{
	/* I *THINK* this is portable, since in_addr.s_addr is
	 * always stored in network byte order[1].
	 *
	 * [1]: See man ip(7)
	 */
	addr->addr.ulong = in->s_addr;
	return 0;
}

int ipv4_to_in_addr(struct ipv4_addr *addr, struct in_addr *in)
{
	/* I *THINK* this is portable, since in_addr.s_addr is
	 * always stored in network byte order[1].
	 *
	 * [1]: See man ip(7)
	 */
	in->s_addr = addr->addr.ulong;
	return 0;
}



































void test_ipv4()
{
	const char *str,*str2;
	char ans[50];
	struct ipv4_range r,a;

	/* Invalids first */
	str = "1.2.256.4";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = "1.2..4";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = ".2.1.4.1";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = "1.2.25";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = "0.0.0.0.0";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = "0.0.0.0/33";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = "0.0.0.0/-1";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = "0.0.0.0/256.0.0.0";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str = "0.0.0.0/0.0.0";
	if (str_to_ipv4(str,&r) == 0) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}

	/* Valids now */

	str = "1.2.3.4";
	a.ip.addr.sep.a = 1;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.ulong = 0xFFFFFFFF;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}


	str = "255.2.3.4";
	a.ip.addr.sep.a = 255;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.ulong = 0xFFFFFFFF;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}


	str = "255.2.3.4";
	a.ip.addr.sep.a = 255;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.ulong = 0xFFFFFFFF;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_to_str(&r, ans, sizeof(ans), IPV4_IP_OPT_SMART_NM) == -1) {
		fprintf(stderr,"fail on %s to str\n",str); 
		exit(1);
	}
	if (strcmp(ans,str) != 0) {
		fprintf(stderr,"fail on %s to str (ans=%s)\n",str,ans);
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}



	str = "255.255.255.255";
	a.ip.addr.sep.a = 255;
	a.ip.addr.sep.b = 255;
	a.ip.addr.sep.c = 255;
	a.ip.addr.sep.d = 255;
	a.nm.addr.ulong = 0xFFFFFFFF;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}



	str = "0.0.0.0";
	a.ip.addr.sep.a = 0;
	a.ip.addr.sep.b = 0;
	a.ip.addr.sep.c = 0;
	a.ip.addr.sep.d = 0;
	a.nm.addr.ulong = 0xFFFFFFFF;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}


	str = "127.0.0.1";
	a.ip.addr.sep.a = 127;
	a.ip.addr.sep.b = 0;
	a.ip.addr.sep.c = 0;
	a.ip.addr.sep.d = 1;
	a.nm.addr.ulong = 0xFFFFFFFF;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}


	str = "1.2.3.4/5.6.7.8";
	a.ip.addr.sep.a = 1;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.sep.a = 5;
	a.nm.addr.sep.b = 6;
	a.nm.addr.sep.c = 7;
	a.nm.addr.sep.d = 8;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_to_str(&r, ans, sizeof(ans), IPV4_IP_WITH_FULL_NM) == -1) {
		fprintf(stderr,"fail on %s to str\n",str); 
		exit(1);
	}
	if (strcmp(ans,str) != 0) {
		fprintf(stderr,"fail on %s to str (ans=%s)\n",str,ans);
		exit(1);
	}
	if (ipv4_legal(&r) == 0) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}


	str = "1.2.3.4/32";
	a.ip.addr.sep.a = 1;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.sep.a = 255;
	a.nm.addr.sep.b = 255;
	a.nm.addr.sep.c = 255;
	a.nm.addr.sep.d = 255;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_to_str(&r, ans, sizeof(ans), IPV4_IP_WITH_SMART_NM) == -1) {
		fprintf(stderr,"fail on %s to str\n",str); 
		exit(1);
	}
	if (strcmp(ans,str) != 0) {
		fprintf(stderr,"fail on %s to str (ans=%s)\n",str,ans);
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}


	str = "1.2.3.4/16";
	a.ip.addr.sep.a = 1;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.sep.a = 255;
	a.nm.addr.sep.b = 255;
	a.nm.addr.sep.c = 0;
	a.nm.addr.sep.d = 0;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}



	str = "1.2.3.4/25";
	a.ip.addr.sep.a = 1;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.sep.a = 255;
	a.nm.addr.sep.b = 255;
	a.nm.addr.sep.c = 255;
	a.nm.addr.sep.d = 128;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_to_str(&r, ans, sizeof(ans), IPV4_IP_OPT_FULL_NM) == -1) {
		fprintf(stderr,"fail on %s to str\n",str); 
		exit(1);
	}

	if (ipv4_to_str(&r, ans, sizeof(ans), IPV4_IP_OPT_SMART_NM) == -1) {
		fprintf(stderr,"fail on %s to str\n",str); 
		exit(1);
	}
	if (strcmp(ans,str) != 0) {
		fprintf(stderr,"fail on %s to str (ans=%s)\n",str,ans);
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}



	str = "1.2.3.4/2";
	a.ip.addr.sep.a = 1;
	a.ip.addr.sep.b = 2;
	a.ip.addr.sep.c = 3;
	a.ip.addr.sep.d = 4;
	a.nm.addr.sep.a = 192;
	a.nm.addr.sep.b = 0;
	a.nm.addr.sep.c = 0;
	a.nm.addr.sep.d = 0;
	if (str_to_ipv4(str,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (r.ip.addr.ulong != a.ip.addr.ulong ||
			r.nm.addr.ulong != a.nm.addr.ulong) { 
		fprintf(stderr,"fail on %s bad answer\n",str); 
		exit(1);
	}
	if (ipv4_legal(&r) == -1) {
		fprintf(stderr,"fail on legal check of %s\n",str); 
		exit(1);
	}

	str = "192.168.0.1";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.168.0.0/255.255.255.0";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != 0) {
		fprintf(stderr,"fail on comp %s should match %s\n",str,str2); 
		exit(1);
	}

	str = "192.168.1.1";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.168.0.0/255.255.255.0";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != -1) {
		fprintf(stderr,"fail on comp %s shouldn't match %s\n",
				str,str2); 
		exit(1);
	}

	str = "192.168.0.1";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.168.0.0/255.255.255.192";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != 0) {
		fprintf(stderr,"fail on comp %s should match %s\n",str,str2); 
		exit(1);
	}

	str = "192.168.0.63";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.168.0.0/255.255.255.192";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != 0) {
		fprintf(stderr,"fail on comp %s should match %s\n",str,str2); 
		exit(1);
	}

	str = "192.168.0.64";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.168.0.0/255.255.255.192";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != -1) {
		fprintf(stderr,"fail on comp %s shouldn't match %s\n",
				str,str2); 
		exit(1);
	}

	str = "192.168.0.70";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.168.0.64/255.255.255.192";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != 0) {
		fprintf(stderr,"fail on comp %s should match %s\n",str,str2); 
		exit(1);
	}
	str = "192.168.0.128";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.168.0.64/255.255.255.192";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != -1) {
		fprintf(stderr,"fail on comp %s shouldn't match %s\n",
				str,str2); 
		exit(1);
	}

	/* Interesting match */
	str = "192.168.0.128";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.2.0.64/255.0.255.0";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != 0) {
		fprintf(stderr,"fail on comp %s should match %s\n",
				str,str2); 
		exit(1);
	}
	if (ipv4_legal(&r) != -1) {
		fprintf(stderr,"range should be invalid %s\n",str2);
		exit(1);
	}

	/* Interesting match */
	str = "192.168.0.129";
	if (str_to_ipv4_addr(str,&(a.ip)) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	str2 = "192.2.0.130/255.0.255.128";
	if (str_to_ipv4(str2,&r) == -1) {
		fprintf(stderr,"fail on %s\n",str); 
		exit(1);
	}
	if (ipv4_comp(&a.ip, &r) != 0) {
		fprintf(stderr,"fail on comp %s should match %s\n",
				str,str2); 
		exit(1);
	}
	if (ipv4_legal(&r) != -1) {
		fprintf(stderr,"range should be invalid %s\n",str2);
		exit(1);
	}






	fprintf(stderr,"success\n");
	exit(0);

}
